import logging
import os
import pickle  # lazy loaded
from abc import abstractmethod
from typing import Dict, List, Union

from wiki_music.constants import (EXTENDED_TAGS, GREEN, LBLUE, LGREEN,
                                  OUTPUT_FOLDER, RESET, YELLOW)
from wiki_music.utilities import (
    SharedVars, ThreadPool, bracket, count_spaces, list_files,
    normalize_caseless, win_naming_convetion, write_roman, yaml_dump)

from ..lyrics import save_lyrics
from ..tags_io import read_tags, write_tags
from .base import ParserBase

logging.getLogger(__name__)

wnc = win_naming_convetion
nc = normalize_caseless

__all__ = ["ParserInOut"]


class ParserInOut(ParserBase):
    """ Class that is inherited by
    :class:`wiki_music.library.parser.process_page.WikipediaParser`. Takes care
    of outputing and loading information.
    """
    def __init__(self, protected_vars):

        super().__init__(protected_vars=protected_vars)

        self._debug_folder: str = ""

    @abstractmethod
    def info_tracks(self):
        raise NotImplementedError("Call to abstract method")

    @property
    def bracketed_types(self) -> List[str]:  # type: ignore
        """ takes `_bracketed_types` list, populates and returns it.

        See also
        --------
        :func:`wiki_music.utilities.parser_utils.bracket`
            function used to append brackets at the ends of strings in list

        :type: List[str]
        """

        if not self._bracketed_types:
            self._bracketed_types = bracket(self.types)
            return self._bracketed_types
        else:
            return self._bracketed_types

    @property  # type: ignore
    def files(self) -> List[str]:  # type: ignore
        """ Gets list of music files in currently set working directory.

        :type: List[str]

        See also
        --------
        :meth:`reassign_files`

        """

        if len(self._files) < len(self.tracks):
            self.reassign_files()

        return self._files

    @files.setter
    def files(self, files: List[str]):
        self._files = files

    @property
    def debug_folder(self) -> str:
        """Path to debugging folder.

        :type: str
        """

        if not self._debug_folder:
            _win_name = win_naming_convetion(self.album, dir_name=True)
            self._debug_folder = os.path.join(OUTPUT_FOLDER, _win_name)
            os.makedirs(self._debug_folder, exist_ok=True)

        return self._debug_folder

    def list_files(self):
        """Lists files in current working directory.

        See also
        --------
        :func:`wiki_music.utilities.utils.list_files`
            function used for file searching
        :attr:`work_dir`
        """
        self.files = list_files(self.work_dir)

    def reassign_files(self):
        """Searched the current working directory and tries to assign files to
        the tracks.
        """

        wnc = win_naming_convetion
        max_length = len(max(self.tracks, key=len))

        # write data to ID3 tags
        disk_files = (self.work_dir)
        files = []

        print(GREEN + "\nFound files:")
        print(*disk_files, sep="\n")
        print(GREEN + "\nAssigning files to tracks:")

        for i, tr in enumerate(self.tracks):
            self.tracks[i] = tr.strip()

            for path in disk_files:
                f = os.path.split(path)[1]
                if (nc(wnc(tr)) in nc(f)
                        and nc(self.types[i]) in nc(f)):  # noqa E129

                    print(LBLUE + tr + RESET,
                          "-" * (1 + max_length - len(tr)) + ">", path)
                    files.append(path)
                    break
            else:
                print(YELLOW + tr + RESET, "." * (2 + max_length - len(tr)),
                      "Does not have a matching file!")

                files.append(None)

        self.files = files

    def basic_out(self):
        """ Outputs files in three basic formats:\n
        1. pickled version of the downloaded wikipedia page
        2. nicely formated html version of the wikipedia page
        3. plain text version of the wikipedia page
        """

        os.makedirs(self.debug_folder, exist_ok=True)

        # save page object for offline debbug
        fname = os.path.join(self.debug_folder, 'page.pkl')
        if not os.path.isfile(fname):
            with open(fname, 'wb') as f:
                pickle.dump(self.page, f)

        # save formated html to file
        fname = os.path.join(self.debug_folder, 'page.html')
        if not os.path.isfile(fname):
            with open(fname, 'w', encoding='utf8') as f:
                f.write(self.soup.prettify())

        # save html converted to text
        fname = os.path.join(self.debug_folder, 'page.txt')
        if not os.path.isfile(fname):
            with open(fname, 'w', encoding='utf8') as f:
                f.write(self.soup.get_text())

    def disk_write(self):
        """ Saves tracklist and album personnel to disk in plain text formats.
        """

        for i, tracklist in enumerate(self.tracklist_2_str(to_file=True), 1):
            fname = os.path.join(self.debug_folder, f"tracklist_{i}.txt")
            with open(fname, "w", encoding="utf-8") as f:
                f.write(tracklist)

        # save found personel to file
        with open(os.path.join(self.debug_folder, "personnel.txt"),
                  "w",
                  encoding="utf8") as f:

            f.write(self.personnel_2_str())

    def tracklist_2_str(self, to_file=True) -> list:
        """ Convert tracklist to string to print out or write to disk.

        Parameters
        ----------
        to_file: bool
            if False and the tracklist is to be printed to console, highlight
            headers to make tracklist more readable

        Returns
        -------
        str
            nicely formated string representation of tracklist
        """
        def _set_color():
            if to_file:
                return [""] * 3
            else:
                return GREEN, LGREEN, RESET

        # compute number of spaces, wrong mypy detection
        spaces, length = count_spaces(self.tracks,
                                      self.bracketed_types)  # type: ignore

        G, LG, R = _set_color()

        # convert to string
        tracklists = []
        for j, (ds, hd) in enumerate(zip(self.disks, self.header)):
            s = ""
            s += f"{G}\n{j + 1}. Disk:{R} {ds}"
            if len(hd) >= 2:
                s += f" {LG}{hd[0]} {hd[1]}"
            if len(hd) >= 3:
                s += f"{' '*(length + len(hd[1]))}{hd[2]}"
            if len(hd) >= 4:
                s += f", {hd[3]}"
            s += f"{R}\n"

            for i in range(self.disk_sep[j], self.disk_sep[j + 1]):
                s += (f"{self.numbers[i]:>2}. {self.tracks[i]} "
                      f"{self.bracketed_types[i]}{spaces[i]} "
                      f"{', '.join(self.artists[i] + self.composers[i])}\n")

                for k, (sbtr, sbtp) in enumerate(
                        zip(self.subtracks[i], self.sub_types[i])):
                    s += (f"  {write_roman(k + 1)}. {sbtr} {sbtp}\n")

            tracklists.append(s)

        return tracklists

    def print_tracklist(self):
        """ Prints tracklist to console.

        See also
        --------
        :func:`tracklist_2_str`
        """

        print("\n".join(self.tracklist_2_str(to_file=False)))

    def personnel_2_str(self):
        """ Convert album personnel to string to print out or write to disk.

        Returns
        -------
        str
            nicely formated string representation of personnel
        """

        s = ""
        if not self.personnel:
            s += "---\n"

        for pers, app in zip(self.personnel, self.appearences):

            if app:
                s += pers + " - "
                temp = 1000

                for k, a in enumerate(sorted(app)):
                    for j, _ in enumerate(self.disk_sep[:-1]):

                        if (a >= self.disk_sep[j]
                                and a < self.disk_sep[j + 1]):  # noqa E129

                            if j != temp:
                                s += f"{self.disks[j]}: {self.numbers[a]}"
                                temp = j
                            else:
                                s += self.numbers[a]

                            if k != len(app) - 1:
                                s += ", "
                            break
                    else:
                        pass

                s += u'\n'
            else:
                s += pers + u'\n'

        return s

    def data_to_dict(self
                     ) -> List[Dict[str, Union[str, int, bytearray, list]]]:
        """ Converts parser data to list of dictionaries. If yaml_dump is
        enabled list is written to 
        
        Warnings
        --------
        This class is not ment to be instantiated, only inherited.

        See also
        --------
        :const:`wiki_music.constants.tags.EXTENDED_TAGS`
            list of tags that are written to each dictionary
        :attr:`wiki_music.utilitie.sync.SharedVars.yaml_dump`

        Returns
        -------
        List[Dict[str, Union[str, int, bytearray, list]]]
            each dictionary in list represents tags of one song
        """

        dict_data = []
        for i, _ in enumerate(self.tracks):

            tags = dict()

            for t in EXTENDED_TAGS:
                attr = getattr(self, t)
                if isinstance(attr, list):
                    tags[t] = attr[i]
                else:
                    tags[t] = attr

            dict_data.append(tags)

        if SharedVars.write_json:
            yaml_dump(dict_data, self.work_dir)
            print("Saved YAML file")

        return dict_data

    def write_tags(self) -> bool:
        """ Write tags to coresponding files. Writing is done in a parallel
        maner to speed things up.

        See also
        --------
        :func:`wiki_music.library.tags_io.write_tags`
            function that handles tag writing
        :meth:`data_to_dict`
            this method prepares tags data in suitable format for writing
        :func:`wiki_music.utilities.parser_utils.ThreadPool`
            class that handles paralelism

        Returns
        -------
        bool
            If writing was successfull return true value
        """

        if not any(self.files):
            return False
        else:
            ThreadPool(target=write_tags,
                       args=[(data, ) for data in self.data_to_dict()]).run()

            return True

    def save_lyrics(self):
        """Function that calls lyricsfinder to search for and save lyrics for
        all tracks.

        See also
        --------
        :func:`wiki_music.library.lyrics.save_lyrics`
            function that handles lyrics finding and saving

        Note
        ----
        lyrics search is controled by
        :attr:`wiki_music.utilities.sync.SharedVars.write_lyrics`
        """

        if SharedVars.write_lyrics:
            self.lyrics = save_lyrics(self.tracks, self.types, self.band,
                                      self.album, self.GUI)
        else:
            self.lyrics = [""] * len(self)

    def read_files(self):
        """ Read tags from files in working directory.

        See also
        --------
        :meth:`list_files`
            method that takes care of file searching
        :func:`wiki_music.library.tags_io.read_tags`
            function that thandles tag reading
        """

        # initialize variables
        self.__init__(protected_vars=False)

        self.list_files()

        for fl in self.files:

            for key, value in read_tags(fl).items():

                # TODO need more elegant way to avoid this difference
                # between read tags and parser attributes
                if key == "COMMENT":
                    continue

                if isinstance(getattr(self, key), list):
                    getattr(self, key).append(value)
                else:
                    setattr(self, key, value)

        if self.files:
            # look for aditional artists in brackets behind track names and
            # complete artists names again with new info
            self.info_tracks()

            self.log.info("Files loaded sucesfully")
        else:
            self.album = ""
            self.band = ""
            self.selected_genre = ""
            self.release_date = ""

            self.log.info("No music files to Load")



"""Module for handling m4a tags."""

import logging
from collections import OrderedDict
from typing import Dict, Union, cast

from mutagen.mp4 import MP4, MP4Cover, MP4MetadataError

from .tag_base import TagBase

log = logging.getLogger(__name__)
log.debug("loading m4a module")


class TagM4a(TagBase):
    """A low level implementation of tag handling for m4a files. """
    __doc__ += TagBase.__doc__  # type: ignore

    _map_keys = OrderedDict([
        ("\xa9alb", "ALBUM"),
        ("aART", "ALBUMARTIST"),
        ("\xa9ART", "ARTIST"),
        ("\xa9cmt", "COMMENT"),
        ("\xa9wrt", "COMPOSER"),
        ("\xa9day", "DATE"),
        ("disk", "DISCNUMBER"),
        ("\xa9gen", "GENRE"),
        ("\xa9lyr", "LYRICS"),
        ("\xa9nam", "TITLE"),
        ("trkn", "TRACKNUMBER"),
        ("covr", "COVERART")]
    )

    def _open(self, filename: str):
        """Function reading m4a file to mutagen.mp4.MP4 class. """

        try:
            self._song = MP4(filename=filename)
        except MP4MetadataError:
            print("Cannot read MP4 tags")

    def _read(self) -> Dict[str, Union[str, bytearray]]:

        tags = dict()
        for key, value in self._map_keys.items():  # pylint: disable=no-member
            try:
                tag = self._song.tags[key]
                if isinstance(tag, list):
                    tag = tag[0]
                if value in ("DISCNUMBER", "TRACKNUMBER"):
                    tag = tag[0]
                if value != "COVERART":
                    tag = str(tag).strip()
            except KeyError:
                tag = ""
            finally:
                tags[value] = tag

        return tags

    def _write(self, tag: str, value: Union[str, bytearray]):

        if tag in ("DISCNUMBER", "TRACKNUMBER"):
            value = [[int(value), 0]]  # type: ignore
        elif tag == "COVERART":
            value = [MP4Cover(value, imageformat=MP4Cover.FORMAT_JPEG)]  # type: ignore

        self._song.tags[self.reverse_map[tag]] = value



""" Constants used by :mod:`wiki_music.gui_lib` """
from os import path
from typing import Tuple

from .paths import module_path

__all__ = ["GUI_HEADERS", "ASPECT_RATIOS", "SPLIT_HEADERS", "MAIN_WINDOW_UI",
           "COVER_ART_SEARCH_UI", "COVER_ART_EDIT_UI", "API_KEY_MESSAGE"]

#: defines headers for tracklist table in GUI
GUI_HEADERS: Tuple[str, ...] = ("TRACKNUMBER", "TITLE", "TYPE", "ARTIST",
                                "COMPOSER", "DISCNUMBER", "LYRICS", "FILE")
#: tuple of permited aspect ratios for cover art cropping
ASPECT_RATIOS: Tuple[str, ...] = ("Free", "Preserve", "16:9", "4:3", "3:2",
                                  "1:1")
#: define table columns that should be splited because they are list types in
#: parser
SPLIT_HEADERS: Tuple[str, ...] = ("TYPE", "ARTIST", "COMPOSER")
#: path to Main window Ui file
MAIN_WINDOW_UI: str = path.join(module_path(), "ui", "MainWindow.ui")
#: path to Cover art search dialog Ui file
COVER_ART_SEARCH_UI: str = path.join(module_path(), "ui", "CoverArtSearch.ui")
#: path to Cover art editor Ui file
COVER_ART_EDIT_UI: str = path.join(module_path(), "ui", "CoverArtEdit.ui")
#: description text for google api key getting
API_KEY_MESSAGE: str = (
    "To enhance lyrics searching capabilities Wiki Music uses google\n"
    "custom search. To use it you have to get your own API key. It is\n"
    "recomended but not necessary. You will be redirected to page in\n"
    "browser where you will log in with your google account and create\n"
    "a new project. After that go to the API section, search for\n"
    "Custom Search API and enable it. Finally go to credentials\n"
    "section and create new API key.\n"
)



import logging
import os
from typing import List, NoReturn, Tuple, Dict
import queue

from PIL import Image

from wiki_music.constants.paths import OFFLINE_DEBUG_IMAGES  # pylint: disable=import-error
from wiki_music.utilities.utils import list_files

log = logging.getLogger(__name__)

log.info("Loaded Offline google images download")


class googleimagesdownload:
    """ Offline version imitating google images download. Main puprose is
    offline testing.

    Attributes
    ----------
    stack: queue.Queue
        a FIFO stack that contains all the downloaded images, the limit is set
        to 5. Then the downloading is paused until items from the queue are
        consumed.
    max: int
        maximum number of file that are loadable from directory
    files: List[str]
        list of image file paths
    """

    def __init__(self) -> None:
        self.stack = queue.Queue()
        self._exit: bool = False
        self._files: List[str] = []

    def download(self, arguments: dict):
        """ Start reding images from files.

        Parameters
        ----------
        arguments: dict
            dictionary of arguments, essentialy it is not needed. It is
            included only to maintain simillarity with original version API
        """

        dim: Tuple[int, int]
        size: float
        thumb: bytes

        successCount: int = 0
        errorCount: int = 0

        print(f"\nItem no.: 1 --> Item name = {arguments['keywords']}")
        print("Evaluating...")

        for f in self.files:

            try:
                dim = Image.open(f).size
                size = os.path.getsize(f)

                with open(f, "rb") as infile:
                    thumb = infile.read()
            except Exception as e:
                print(e)
                errorCount += 1
            else:
                self.stack.put({"thumb": thumb, "dim": (size, dim), "url": f})
                successCount += 1
                print(f"Completed Image Thumbnail ====> {successCount}. {f}")
            finally:
                if self._exit:
                    print("Album art search exiting ...")
                    return

        print(f"\nErrors: {errorCount}\n")

    def close(self):
        """ Stop downloading images. """
        self._exit = True

    @property
    def max(self) -> int:
        """ Returns maximum number of loadable images. Needed to set progresbar
        in GUI. The value is cached for later use.

        See also
        --------
        :func:`wiki_music.utilities.utils.list_files`
            to see list of suported files

        Returns
        -------
        int
            number of image files
        """

        return len(self.files)

    @property
    def files(self) -> List[str]:
        """ List of image files to load in direstory.

        See also
        --------
        :func:`wiki_music.utilities.utils.list_files`
            to see list of suported files
        :const:`wiki_music.constants.paths.OFFLINE_DEBUG_IMAGES`
            directory that is searched for images

        Returns
        -------
        List[str]
            list of paths to image files
        """

        if not self._files:
            self._files = list_files(OFFLINE_DEBUG_IMAGES, file_type="image",
                                     recurse=True)

        return self._files



"""wiki_music CLI entry point."""

import logging
import signal
import sys

from wiki_music.constants.colors import GREEN, RESET
from wiki_music.library import WikipediaRunner
from wiki_music.utilities import SharedVars, input_parser, set_log_handles


# add signal handler to exit gracefully upon Ctrl+C
def signal_handler(sig, frame):
    print("\nAborting by user request...")
    sys.exit()


def main():

    signal.signal(signal.SIGINT, signal_handler)

    (SharedVars.write_json, SharedVars.offline_debbug, only_lyrics,
     album, band, work_dir, with_log, debug) = input_parser()

    if debug:
        set_log_handles(logging.DEBUG)
    else:
        set_log_handles(logging.WARNING)

    if not album:
        print(GREEN + "Enter album name: " + RESET, end="")
        album = str(input())
    if not band:
        print(GREEN + "Enter band name: " + RESET, end="")
        band = str(input())
    print(RESET)

    parser = WikipediaRunner(GUI=False)

    parser.album = album
    parser.band = band
    parser.work_dir = work_dir
    parser.with_log = with_log
    parser.list_files()

    if only_lyrics:
        parser.run_lyrics()
    else:
        parser.run_wiki()


if __name__ == "__main__":
    main()



""" Fancy wrapper function used in whole package. """

import logging
import os
import time  # lazy loaded
from functools import wraps
from threading import current_thread
from typing import TYPE_CHECKING, Any, Callable, List, NoReturn, Union

from wiki_music.constants import LOG_DIR

from .exceptions import *
from .sync import SharedVars

logging.getLogger(__name__)

if TYPE_CHECKING:
    from logging import Logger
    from threading import Lock


def exception(logger: "Logger", show_GUI: bool = True) -> Callable:
    """
    A decorator that wraps the passed in function and logs exceptions should
    one. Messages are sent to gui through SharedVars and to logger.

    Warnings
    --------
    Use with caution, only on critical parts of the code and not in early
    development stage as it will hide code errors

    Parameters
    ----------
    logger: logging.Logger
        logger instance whih will be recording occured errors
    show_GUI: bool
        whether to show the exception message in GUI messagebox

    Returns
    -------
    Callable
        wrapped callable which will not crash app when it raises error
    """
    def real_wrapper(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except Exception as e:
                logger.exception(f"Unhandled golbal exception: {e}")
                if show_GUI:
                    SharedVars.exception(f"Unhandled golbal exception: {e}")

        return wrapper
    return real_wrapper


def synchronized(lock: "Lock") -> Callable:
    """ Synchronization decorator. Syncs all callables wrapped by this
    decorator which are passed the same lock

    Warnings
    --------
    Do not use on computationally heavy functions, as this could cause GUI
    freezing. When dealing with such functions always lock only single
    variables
    
    Parameters
    ----------
    lock: threading.Lock
        lock instance

    Returns
    -------
    Callable
        callable synchronized with passed lock
    """

    def real_wrapper(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args, **kwargs):
            lock.acquire()
            try:
                return function(*args, **kwargs)
            finally:
                lock.release()
        return wrapper
    return real_wrapper


def warning(logger: "Logger", show_GUI: bool = True) -> Callable:
    """
    A decorator that wraps the passed in function and logs AttributeErrors
    to logger warning and Exceptions to logger exceptions. Messages are sent to
    gui through SharedVars and to logger.

    Warnings
    --------
    Use with caution, only on critical parts of the code and not in early
    development stage as it will hide code errors

    Parameters
    ----------
    logger: logging.Logger
        logger instance whih will be recording occured errors
    show_GUI: bool
        whether to show the warning message in GUI messagebox

    Returns
    -------
    Callable
        wrapped callable which will not crash app when it raises error
    """

    def real_wrapper(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except (NoTracklistException, NoReleaseDateException,
                    NoGenreException, NoCoverArtException,
                    NoNames2ExtractException, NoContentsException,
                    NoPersonnelException, Mp3tagNotFoundException) as e:
                logger.warning(e)
                if show_GUI:
                    SharedVars.warning(e)

        return wrapper
    return real_wrapper


class Timer:
    """ Timing context manager. measures execution time of a function.

    Parameters
    ----------
    function_name: str
        name of the timed function that will be visible in log

    Attributes
    ----------
        start: float
            execution start time
        end: float
            excecution end time
        path: string
            string contining path to file where results are logged
    """

    start: float
    end: float

    def __init__(self, function_name: str) -> None:
        self.function_name = function_name
        self.path = os.path.join(LOG_DIR, "timing_stats.log")
        self.start = 0
        self.end = 0

        if not os.path.isfile(self.path):
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            with open(self.path, "w") as f:
                f.write("")

    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()

        with open(self.path, "a") as f:
            f.write(f"{current_thread().name:15} --> {self.function_name:30}"
                    f"{(self.end - self.start):8.4f}s\n")


def time_methods(function: Callable) -> Callable:
    """ A decorator that wraps the passed in function and measures execution
    time.

    Parameters
    ----------
    function: Callable
    """
    @wraps(function)
    def wrapper(*args, **kwargs) -> Callable:
        with Timer(function.__name__):
            return function(*args, **kwargs)

    return wrapper

# TODO why exceptions are thrown for static methods??
def for_all_methods(decorator: Callable[[Any], Callable],
                    exclude: List[str] = []) -> Callable:
    """ Decorates class methods, except the ones in excluded list.

    Warnings
    --------
    staticmetods of a class and inner classes must be excluded from decorating
    ofherwise exception is thrown.

    References
    ----------
    https://stackoverflow.com/questions/6307761/how-to-decorate-all-functions-of-a-class-without-typing-it-over-and-over-for-eac
    
    Parameters
    ----------
    decorator: Callable
        Callable which will be used to decorate class methods
    exclude: List[str]
        list of method names to exclude from decorating
    """

    @wraps(decorator)
    def decorate(cls: Callable) -> Callable:
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate



""" This module containns the whole parser with all the inherited subclasses.
Class :class:`WikipediaParser` has complete functionallity but its methods need
to be called in the correst order to give sensible results.
"""

import logging
import re  # lazy loaded
from os import path
from threading import Thread
from typing import List

import datefinder  # lazy loaded
import fuzzywuzzy.fuzz as fuzz  # lazy loaded
import fuzzywuzzy.process as process  # lazy loaded

from wiki_music.constants import FILES_DIR
from wiki_music.constants.parser_const import *
from wiki_music.utilities import (NLTK, SharedVars, caseless_contains,
                                  complete_N_dim, delete_N_dim, flatten_set,
                                  for_all_methods, get_image, normalize,
                                  normalize_caseless, replace_N_dim,
                                  time_methods, warning)
from wiki_music.utilities.exceptions import *

from .base import ParserBase
from .extractors import DataExtractors
from .in_out import ParserInOut
from .preload import WikiCooker

nc = normalize_caseless
log = logging.getLogger(__name__)

NLTK.run_import(log)  # imports nltk in separate thread

log.debug("parser imports done")

__all__ = ["WikipediaParser"]


@for_all_methods(time_methods)
class WikipediaParser(DataExtractors, WikiCooker, ParserInOut):
    """ Class for parsing the wikipedia page and extracting tags data from it.

    Warnings
    --------
    Most parser methods are designed to fail gracefully so the extractions can
    proceed even when some subset of it failed. This has a dark side because
    it hides errors!!! All warning decorated methods are resilient to any
    exception defined in :mod:`wiki_music.utilities.exceptions` .

    References
    ----------
    https://www.crummy.com/software/BeautifulSoup/bs4/doc/: used to parse the
    wikipedia page

    Parameters
    ----------
    protected_vars: bool
        defines if certain variables should be initialized by __init__ method
        or not
    """   

    def __init__(self, protected_vars: bool = True) -> None:

        log.debug("init parser")

        WikiCooker.__init__(self, protected_vars=protected_vars)
        ParserInOut.__init__(self, protected_vars=protected_vars)

        log.debug("init parser done")

    @warning(log)
    def get_release_date(self):
        """ Gets album release date from information box in the
        top right corner of wikipedia page. Populates:attr:`wiki_music.DATE`

        Raises
        ------
        :exc:`utilities.exceptions.NoReleaseDateException`
            raised if no release date was extracted
        """

        dates = self.sections["infobox"].find(class_="published")

        if dates:
            dates = datefinder.find_dates(dates.get_text())
            date_year = [d.strftime('%Y') for d in dates]
            self.release_date = list(set(date_year))[0]
        else:
            self.release_date = ""
            raise NoReleaseDateException

    @warning(log)
    def get_genres(self):
        """ Gets list of album genres from information box in the
        top right corner of wikipedia page. If found genre if only one then
        assigns is value to :attr:`GENRE`

        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoGenreException`
            if no genres could be extracted from page
        """

        genres = self.sections["infobox"].find(class_="category hlist")

        if genres:
            self.genres = [g.string for g in genres.find_all(href=WIKI_GENRES,
                                                             title=True)]
        else:
            self.genres = []
            raise NoGenreException

        # auto select genre if only one was found
        if len(self.genres) == 1:
            self.selected_genre = self.genres[0]

    @warning(log)
    def get_cover_art(self):
        """ Gets album cover art information box in the top right corner
        of wikipedia page. Runs in a separate thread because the cover art
        data is not used by parser in any way, so it can be downloaded in
        the background. Populates :attr:`COVERART`

        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoCoverArtException`
            if cover art url could not be found
        """

        # self.cover_art is not protected by lock, parser should have
        # more than enough time to get it before GUI requests it
        def cover_art_getter():

            """
            for child in self.sections["infobox"].children:
                if child.find("img") is not None:
                    image = child.find("img")
                    image_url = f"https:{image['src']}"

                    self.cover_art = get_image(image_url)
                    break
            """
            for img in self.sections["infobox"].find_all("img", src=True,
                                                         alt=True):
                if fuzz.token_set_ratio(img["alt"], self.album) > 90:
                    break

            if img:
                self.cover_art = get_image(f"https:{img['src']}")
            else:
                with open(path.join(FILES_DIR, "Na.jpg"), "rb") as f:
                    self.cover_art = bytearray(f.read())
                raise NoCoverArtException("Couldn't extract cover art from "
                                          "wikipedia page")

        Thread(target=cover_art_getter, name="CoverArtGetter").start()

    def get_composers(self):
        """ Extracts composers from wikipedia page. Employs complex logic.
        First Person named entities are extracted by nltk. Then merges them
        with composers. After that uses this list of names to try to guess
        composers and coresponding tracks from short text above the table.

        See also
        --------
        :meth:`get_personnel`
            this method should run first because it populates the
            :attr:`personnel` used by this method

        Warnings
        --------
        This method is not as robust as it should be. It fails for many
        types of formating.
        """

        def check_name_complete(name, text):
            """ Checks if name retrieved fromm text is complete or only part by
            checking the following words in text.

            Warnings
            --------
            This method is not really dependable, has quite poor results
            """

            # find out if next word starts with capital
            pos_1 = text.find(normalize(name)) + len(name) + 1
            # safety measure for when we hit the string end or
            # don´t find any entries
            if pos_1 > len(text) - 1 or pos_1 == -1:
                pos_1 = len(text) - 1
            # find out if the preceeding word starts with capital
            pos_2 = text.rfind(" ", 0, text.find(normalize(name)) - 1) + 1
            # if yes add it to name
            # useful if NLTK finds only one part of name

            if text[pos_1].isupper():
                surname = text[pos_1: text.find(" ", pos_1 + 1)]
                if surname not in name:
                    return name + " " + surname
            elif text[pos_2].isupper() and text[pos_2] != ".":
                surname = text[pos_2: text.find(" ", pos_2 + 1)]
                if surname not in name:
                    return name + " " + surname
            else:
                return name

        # extract names from text using NLTK
        self.extract_names()

        self.NLTK_names = set(self.NLTK_names + self.personnel)

        # get the short comment above the table which is marked as html
        # paragraph with <p>...</p>
        if self.sections["track_listing"][0].name == "p":
            html = self.sections["track_listing"][0].get_text()
        else:
            html = ""

        # find NLTK names that occure in sentence after specified phrases
        # split to sentences
        sentences = []

        for sentence in html.split("."):
            for d in DELIMITERS:
                if re.search(d, sentence, re.IGNORECASE):
                    sentences.append(sentence)

        # extract composers from sentences
        for sentence in sentences:
            parts = sentence.split("except")

            # find which track are affected by except
            index = []
            if len(parts) == 2:
                for i, tr in enumerate(self.tracks):
                    if caseless_contains(tr, parts[1]):
                        index.append(i)

            # assign composers to tracks
            for name in self.NLTK_names:

                if len(parts) == 2:
                    if caseless_contains(name, parts[1]):
                        n = check_name_complete(name, parts[1])

                        for ind in index:
                            self.composers[ind].append(n)

                if caseless_contains(name, parts[0]):
                    n = check_name_complete(name, parts[0])
                    for i, comp in enumerate(self.composers):
                        if i not in index:
                            if n not in comp:
                                self.composers[i].append(n)

    @warning(log)
    def get_contents(self):
        """ Extract page contets from keys in :attr:`sections` dictionary.
        
        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoContentsException`
            if no contents were retrieved
        """

        # the last element is infobox which was added manually
        self.contents = [s.replace("_", " ").capitalize()
                         for s in self.sections.keys()][:-1]

        if len(self.contents) == 0:
            raise NoContentsException

    @warning(log)
    def get_personnel(self):
        """ Extract personnel from wikipedia page sections defined by:
        :const:`wiki_music.constants.parser_const.PERSONNEL_SECTIONS` then
        parse these entries for additional data like apperences on tracks.

        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoPersonnelException`
            if no table or list with personnel was found
        """

        personnel = []
        appearences = []
        
        for s in PERSONNEL_SECTIONS:

            if s not in self.sections:
                continue

            for html in self.sections[s]:
                # if the toplevel tag is the list itself
                if html.name in ("ul", "ol"):
                    personnel.extend(self._html2python_list(html))
            
                # if the list is nested inside some other tags
                for h in html.find_all(["ul", "ol"]):
                    personnel.extend(self._html2python_list(h))

        if len(personnel) == 0:
            raise NoPersonnelException

        for i, person in enumerate(personnel):

            # make space in appearences
            appearences.append([])

            # remove reference
            person = re.sub(r"\[ *\d+ *\]", "", person)
            
            # split to person and the rest
            try:
                person, appear = re.split(r" \W | as ", person, 1, flags=re.I)
            # if no delimiter is present skip to next
            except ValueError:
                continue

            # find references to song numbers
            for app in re.findall(r"\d{1,2}-\d{1,2}|\d{1,2}", appear):
                # when we find 3-6
                if "-" in app:
                    start, stop = [int(a) for a in app.split("-")]
                    a = [r for r in range(start, stop + 1)]
                    appearences[i].extend(a)
                # simple number
                else:
                    appearences[i].append(int(app))

            # find references to song names
            for j, t in enumerate(self.tracks):
                if re.match(t, appear, re.I):
                    appearences[i].append(j)

            # ensure no duplicates
            appearences[i] = list(set(appearences[i]))
            appearences[i] = [a - 1 for a in appearences[i]]
            personnel[i] = person

            #print(person, "|", appear, "|", appearences[i])               

        self.personnel = personnel
        self.appearences = appearences

    @warning(log)
    def _get_tracks(self):
        """ Method that attempts to extract tracklist from table or list
        format on the wikipedia page.

        See also
        --------
        :meth:`_from_table`
            used to parse tracklist in htlm table
        :meth:`_from_list`
            used to parse tracklist in html list
        :meth:`process_tracks`
            method called to parse raw extracted table and get
            song numbers, artists, composers ...

        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoTracklistException`
            raised if no tracklist in any format was found
        """

        tables = self.soup.find_all("table", class_="tracklist")

        if tables:
            data = self._from_table(tables)
        else:
            try:
                tables = []
                for s in self.sections["track_listing"]:
                    if s.name in ("ul", "ol"):
                        tables.append(s)
                    else:
                        tables.extend(s.find_all(["ul", "ol"]))
            except AttributeError as e:
                msg = (f"No tracklist found!\nURL: {self.url}\nprobably "
                       f"doesn´t belong to album: {self.album} by {self.band}")
                SharedVars.warning(msg)
                raise NoTracklistException(msg)
            else:
                data = []
                for t in tables:
                    data.extend(self._from_list(t))

        self.process_tracks(data)

    def process_tracks(self, data: List[List[str]]):
        """ Process raw extracted list of tables with trackist for
        track details. """

        self.disk_sep.append(0)

        index = 1
        for CD in data:

            self.disks.append([f"{self.album} CD {index}", len(self)])
            index += 1
            for song in CD:
                if "no" in nc(song[0]):
                    # table header - possibly use in future
                    # posladny stlpec je length a ten väčšinou netreba
                    if "length" in nc(song[-1]):
                        self.header.append(song[:-1])
                    else:
                        self.header.append(song)
                elif re.match(ORDER_NUMBER, song[0]):
                    self.numbers.append(song[0].replace(".", ""))
                    tmp1, tmp2 = self._get_track(song[1])
                    self.tracks.append(tmp1)
                    self.subtracks.append(tmp2)

                    if len(song) > 3:
                        self.artists.append([])
                        self.composers.append([])

                        # all columns between track and track length belong to
                        # artists we assign them to artists or composers
                        # based on label in header
                        for i in range(2, len(song) - 1):
                            a = self._get_artist(song[i])
                            try:
                                is_composer = process.extractOne(
                                    self.header[-1][i], COMPOSER_HEADER,
                                    score_cutoff=90, scorer=fuzz.ratio)
                            except:
                                self.artists[-1].extend(a)
                            else:
                                if is_composer:
                                    self.composers[-1].extend(a)
                                else:
                                    self.artists[-1].extend(a)
                    else:
                        self.artists.append([])
                        self.composers.append([])
                elif "total" in nc(song[0]):
                    # total length summary
                    pass
                else:
                    # if all else passes than line must be disc title
                    self.disks[-1] = [song[0], len(self)]

        # bonus track are sometimes marked as another disc
        disks_filtered = [d for d in self.disks if "bonus" not in nc(d[0])]

        self.disk_sep = [i[1] for i in disks_filtered]
        self.disk_sep.append(len(self))
        self.disks = [i[0] for i in disks_filtered]

        # assign disc number to tracks
        for i, _ in enumerate(self.tracks):
            for j, _ in enumerate(self.disk_sep[:-1]):
                if self.disk_sep[j] <= i and i < self.disk_sep[j + 1]:
                    self.disc_num.append(j + 1)

    def info_tracks(self):
        """ Parse track names for aditional information like
        artist, composer, type... . Also get rid of useless strings like
        bonus track, featuring... . These informations are assumed to be
        enclosed in brackets behind the track name.
        """

        self.types = []
        self.sub_types = []

        comp_flat = flatten_set(self.composers)

        # hladanie umelcov ked su v zatvorke za skladbou
        # + zbavovanie sa bonus track a pod.
        for i, tr in enumerate(self.tracks):

            self.types.append("")
            self.sub_types.append([])

            start_list = [m.start() for m in re.finditer(r'\(', tr)]
            end_list = [m.start() for m in re.finditer(r'\)', tr)]

            for start, end in zip(start_list, end_list):

                self.tracks[i] = re.sub(TO_DELETE, "", tr, re.I)

                artist = re.sub("[,:]", "", tr[start + 1:end])
                artist = re.split(r",|\/|\\", artist)

                for art in artist:
                    for un in UNWANTED:
                        art = re.sub(un, "", art, flags=re.IGNORECASE)

                    # TODO all these maybe should be in the UNWANTED loop??
                    # check against additional personnel
                    for person in self.personnel:
                        if fuzz.token_set_ratio(art, person) > 90:
                            self.artists[i].append(person)
                            self.tracks[i] = self._cut_out(tr, start, end)

                    # check agains composers
                    for comp in comp_flat:
                        if fuzz.token_set_ratio(art, comp) > 90:
                            self.artists[i].append(comp)
                            self.tracks[i] = self._cut_out(tr, start, end)

                    # check if instrumental, ...
                    _type, score = process.extractOne(
                        art, DEF_TYPES, scorer=fuzz.token_sort_ratio)
                    if score > 90:
                        self.types[i] = art  # _type
                        self.tracks[i] = self._cut_out(tr, start, end)

            if self.subtracks:
                for j, sbtr in enumerate(self.subtracks[i]):

                    self.sub_types[i] = [""] * len(sbtr)

                    start_list = [m.start() for m in re.finditer(r'\(', sbtr)]
                    end_list = [m.start() for m in re.finditer(r'\)', sbtr)]

                    for start, end in zip(start_list, end_list):

                        art = re.sub("[,:]", "", sbtr[start + 1:end]).split()

                        for a in art:
                            # check against additional personnel
                            for person in self.personnel:
                                if fuzz.token_set_ratio(a, person) > 90:
                                    self.artists[i].append(person)
                                    sbtr = self._cut_out(sbtr, start, end)
                                    self.subtracks[i][j] = sbtr

                            # check if instrumental, ...
                            _type, score = process.extractOne(
                                a, DEF_TYPES, scorer=fuzz.token_sort_ratio)
                            if score > 90:
                                self.sub_types[i][j] = _type
                                sbtr = self._cut_out(sbtr, start, end)
                                self.subtracks[i][j] = sbtr

    def complete(self):
        """ Recursively traverses: :attr:`composers`, :attr:`artists` and
        :attr:`personnel` and checks each name with each if some is found to be
        incomplete then it is replaced by longer version from other list.
        """ 

        to_complete = (self.composers, self.artists, self.personnel)
        delete: list = ["", " "]

        # complete everything with everything
        for to_replace in to_complete:
            for to_find in to_complete:
                complete_N_dim(to_replace, to_find)

        # sort artist alphabeticaly
        if self.artists:
            for a in self.artists:
                a.sort()

        for tc in to_complete:
            delete_N_dim(tc, delete)

        # get rid of feat., faeturing ...
        for tc in to_complete:
            for un in UNWANTED:
                replace_N_dim(tc, un)

        # get rid of artists duplicates
        for tc in to_complete:
            for i, t in enumerate(tc):
                if isinstance(t, list):
                    tc[i] = sorted(list(set(t)))

    def merge_artist_personnel(self):
        """ Assigns personnel which have appearences specified to coresponding
        list of track artists.
        """

        for person, appear in zip(self.personnel, self.appearences):
            for a in appear:
                self.artists[a].append(person)

    def merge_artist_composers(self):
        """ Move all artists to composers list. This is done or left out based
        on user input. """

        for i, (c, a) in enumerate(zip(self.composers, self.artists)):
            self.composers[i] = sorted(list(filter(None, set(c + a))))

        self.artists = [""] * len(self.composers)

    @warning(log, show_GUI=False)
    def extract_names(self):
        """ Used nltk to estract person names from supplied sections of the
        wikipedia page.

        See also
        --------
        :const:`wiki_music.constants.parser_const.PERSONNEL_SECTIONS`
            these sections of the page + track_listing are passed to nltk
            to look for names

        Raises
        ------
        :exc:`wiki_music.utilities.exceptions.NoNames2ExtractException`
            vhen the page doesn't have any of the sections defined in See also
            section. So no text can be provided to nltk. It makes no sense to
            try extraction from others parts of tha page as they are too
            cluttered by hardly classifiable information.
        """

        document = ""
        for key, value in self.sections.items():

            if key in (PERSONNEL_SECTIONS + ("track_listing", )):
                for val in value:
                    document += val.get_text(" ")

        # if none of the sections is present exit method
        if not document:
            raise NoNames2ExtractException

        stop = NLTK.nltk.corpus.stopwords.words('english')

        document = ' '.join([i for i in document.split() if i not in stop])
        sentences = NLTK.nltk.tokenize.sent_tokenize(document)
        sentences = [NLTK.nltk.word_tokenize(sent) for sent in sentences]
        sentences = [NLTK.nltk.pos_tag(sent) for sent in sentences]

        names = []
        for tagged_sentence in sentences:
            for chunk in NLTK.nltk.ne_chunk(tagged_sentence):
                if type(chunk) == NLTK.nltk.tree.Tree:
                    if chunk.label() == 'PERSON':
                        names.append(' '.join([c[0] for c in chunk]))

        names = sorted(list(set(names)))

        # TODO smetimes when two names are separated only by "and", "by"..
        # or such short word the two names are found together as one
        # and the sunsequent completion messes the right names that were found
        """
        # filter incomplete names
        selected_names = []
        for n1 in names:
            name = n1
            for n2 in names:
                if name in n2:
                    name = n2
                else:
                    pass

            selected_names.append(name)
        """

        # filter out already found tracks
        filtered_names = [n for n in names
                          if fuzz.token_set_ratio(n, self.tracks) < 90]

        self.NLTK_names = filtered_names



"""Searches lyrics."""

import logging

from .lyrics import LyricsManager
from .models import *

logging.getLogger(__name__)

extract_lyrics = LyricsManager.extract_lyrics
search_lyrics = LyricsManager.search_lyrics



"""wiki_music GUI entry point."""

import logging
import sys
from threading import current_thread

from wiki_music.gui_lib.main_window import Window
from wiki_music.gui_lib.qt_importer import QApplication
from wiki_music.utilities import input_parser, set_log_handles


def main():

    debug = input_parser()[-1]

    if debug:
        set_log_handles(logging.DEBUG)
    else:
        set_log_handles(logging.INFO)

    current_thread().name = "GuiThread"
    app = QApplication(sys.argv)
    ui = Window(debug=debug)
    ui.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""This module defines all path constants that are used in package."""

import sys
from os import path, remove, makedirs
from appdirs import user_data_dir, user_log_dir


__all__ = ["ROOT_DIR", "DIR_FILE", "LOG_DIR", "MP3_TAG", "OUTPUT_FOLDER",
           "OFFLINE_DEBUG_IMAGES", "FILES_DIR", "DONT_BOTHER_API",
           "GOOGLE_API_URL", "API_KEY_FILE", "module_path"]


def _dir_writable(dir_name: str) -> bool:
    """Test if inpu directory is writable.

    Parameters
    ----------
    dir_name: str
        path to tested directory

    Returns
    -------
    bool
        true if files can be  writen to directory
    """

    test_write = path.join(dir_name, "TEST_WRITE")

    try:
        f = open(test_write, "w")
    except IOError:
        return False
    else:
        f.close()
        remove(test_write)
        return True


def module_path() -> str:
    """Outputs root path of the module. Acounts for changes in path when
    app is frozen.

    Returns
    -------
    str
        string with root module path

    See also
    --------
    :func:`we_are_frozen`
    """

    if hasattr(sys, "frozen"):
        return path.dirname(sys.executable)
    else:
        return path.join(path.dirname(__file__), "..")


#: toplevel directory for file saving
ROOT_DIR: str = module_path()
#: directory containing logs, profiling log and dir with parser output
LOG_DIR: str = path.join(ROOT_DIR, "logs")
#: directory containing images used for offline debugging of image search
OFFLINE_DEBUG_IMAGES: str = path.join(ROOT_DIR, "..", "tests", "offline_debug")
#: directory containing package files like icons etc.
FILES_DIR: str = path.join(ROOT_DIR, "files")
#: path to Mp3tag executable
MP3_TAG: str = r"C:\Program Files (x86)\Mp3tag\Mp3tag.exe"

# get platform specific data dirs if root package dir is not writable
if not _dir_writable(ROOT_DIR):
    ROOT_DIR = user_data_dir(appname="wiki_music", appauthor=None)
    LOG_DIR = user_log_dir(appname="wiki_music", appauthor=None)
    OFFLINE_DEBUG_IMAGES = ""

#: file containing record of last openede directory in gui
DIR_FILE: str = path.join(ROOT_DIR, "data", "last_opened.txt")
#: folder containing human readable text output, of processed wiki page
#: and also pickled version of wikipedia.WikipediaPage
OUTPUT_FOLDER: str = path.join(LOG_DIR, "output")
#: Google API key file
API_KEY_FILE: str = path.join(ROOT_DIR, "files", "google_api_key.txt")
#: file that says if user should be asked to get google API key
DONT_BOTHER_API: str = path.join(ROOT_DIR, "files", "DONT_BOTHER_API")
#: Google API url
GOOGLE_API_URL: str = ("https://console.developers.google.com/projectselector/"
                       "apis/library/customsearch.googleapis.com")




"""Low level tags handling implementation based on mutagen library."""

import logging
from typing import TYPE_CHECKING, Type

from lazy_import import lazy_callable

from wiki_music.utilities import SharedVars, exception

TagMp3 = lazy_callable("wiki_music.library.tags_handler.mp3.TagMp3")
TagFlac = lazy_callable("wiki_music.library.tags_handler.flac.TagFlac")
TagM4a = lazy_callable("wiki_music.library.tags_handler.m4a.TagM4a")

if TYPE_CHECKING:
    from wiki_music.library.tags_handler.tag_base import TagBase

log = logging.getLogger(__name__)


@exception(log)
def File(filename: str) -> Type["TagBase"]:
    """Class factory function which returns coresponding class based on file
    type.

    Note
    ----
    Currently supported types are: mp3, flac, m4a

    Raises
    ------
    NotImplementedError
        if file is not one of supported types

    Returns
    -------
    TagBase
        one of the low level tag handling classes
    """

    if filename.lower().endswith(".mp3"):
        return TagMp3(filename)
    elif filename.lower().endswith(".flac"):
        return TagFlac(filename)
    elif filename.lower().endswith(".m4a"):
        return TagM4a(filename)
    else:
        e = (f"Tagging for {filename.rsplit('.', 1)[1]} files is not "
             f"implemented")
        raise NotImplementedError(e)



"""Module storing class that takes care of downolading wikipedia page."""

import collections  # lazy loaded
import logging
import pickle  # lazy loaded
import queue
import re  # lazy loaded
import sys
import time  # lazy loaded
from os import path
from threading import Lock
from typing import Dict, Optional, Type

import bs4  # lazy loaded
import fuzzywuzzy.fuzz as fuzz  # lazy loaded

import wikipedia as wiki  # lazy loaded
from wiki_music.constants import OUTPUT_FOLDER
from wiki_music.utilities import (
    SharedVars, ThreadWithTrace, for_all_methods, normalize_caseless,
    time_methods)

from .base import ParserBase

log = logging.getLogger(__name__)
nc = normalize_caseless


@for_all_methods(time_methods, exclude=["Preload"])
class WikiCooker(ParserBase):
    """Downloades wikipedia page and convertes it to WikipediaPage object.

    Subsequently important parts of the page are extracted to class attributes.
    Class has the ability to run preload of the page in background tread as the
    user types input values in GUI.

    Warnings
    --------
    This class is not ment to be instantiated, only inherited.

    References
    ----------
    https://www.crummy.com/software/BeautifulSoup/bs4/doc/: used to parse the
    wikipedia page
    https://pypi.org/project/wikipedia/: used to get the wikipedia page

    Parameters
    ----------
    protected_vars: bool
        defines if certain variables should be initialized by __init__ method
        or not

    Attributes
    ----------
    page: wikipedia.WikipediaPage
        downloaded page to be parsed by BeautifulSoup
    soup: bs4.BeautifulSoup
        BeautibulSoup object representing the whole page
    sections: Dict[str, bs4.BeautifulSoup]
        the :attr:`soup` split to page sections indexed by their titles
    """
    def __init__(self, protected_vars: bool) -> None:

        super().__init__(protected_vars=protected_vars)

        self.log.debug("cooker imports")

        if protected_vars:
            self.sections = {}
            self.page = None
            self.soup = None

            # control download and cook status are indexed by na of album
            self.wiki_downloaded: queue.Queue = queue.Queue(maxsize=1)
            self.soup_ready: queue.Queue = queue.Queue(maxsize=1)

        # control
        self.preload_running: bool = False
        self.getter_lock: Lock = Lock()
        self.error_msg: Optional[str] = None

        self._url = ""

        # pass reference of current class instance to subclass
        self.Preload.outer_instance = self

        self.log.debug("cooker imports done")

    # TODO maybe we can move needed methods from outer class to preload so we
    # can have more instances of preload running at once, the results of each
    # instance would then be saved to dictionary indexed by the search name
    # and the Preload would not need to have access to outer namespace which
    # is ugly and dangerous ffrom maintainability perspective
    class Preload:
        """Contolling the preload of wikipedia page.

        It is totally self-contained exposes only start and stop methods.

        Attributes
        ----------
        _preload_thread: :class:`wiki_music.utilities.parser_utils.ThreadWithTrace`
            stopable thread instance so we can kill preload
        outer_instance: :class:`wiki_music.library.process_page.WikipediaParser`
            instance of the outer class so the inner class can manipulate
            its attributes
        """

        _preload_thread: Optional[ThreadWithTrace] = None
        outer_instance: Optional[Type[ParserBase]] = None

        @classmethod
        def start(cls):
            """Starts running the preload, variables are read from outer class.

            Note
            ----
            Currently running preload is stopped before new one is started
            """

            cls.stop()

            # first empty records of previous download
            try:
                cls.outer_instance.wiki_downloaded.get(block=False)
            except queue.Empty:
                pass
            try:
                cls.outer_instance.wiki_downloaded.get(block=False)
            except queue.Empty:
                pass
            cls.outer_instance.error_msg = None

            log.debug(f"Starting wikipedia preload for: "
                      f"{cls.outer_instance.album} by "
                      f"{cls.outer_instance.band}")

            cls._preload_thread = ThreadWithTrace(
                target=cls.outer_instance._preload_run, name="WikiPreload")
            cls._preload_thread.start()

        @classmethod
        def stop(cls):
            """Method that stops currently running preload."""
            if cls.outer_instance.preload_running:

                log.debug(f"Stoping wikipedia preload for: "
                          f"{cls.outer_instance.album} by "
                          f"{cls.outer_instance.band}")
                cls._preload_thread.kill()
                cls._preload_thread.join()

                cls.outer_instance.preload_running = False

    @property
    def url(self) -> str:
        """Holds url of the wikipedia page from which the data was downloaded.

        When offline debugging is enabled, this property holds path to the
        pickle page file.

        :type: str

        See also
        --------
        :meth:`wiki_music.parser.in_out.ParserInOut.basic_out`
            method responsible for saving the pickled page to disk
        :meth:`_from_disk`
            method for loading the pickle page from disk
        """
        if not self._url:
            if SharedVars.offline_debbug:
                self._url = path.join(OUTPUT_FOLDER, self.album, "page.pkl")
            else:
                self._url = str(self.page.url)  # type: ignore

        return self._url

    # TODO doesn't work without GUI
    def terminate(self, message: str):
        """Send message to GUI to ask user if he wishes to terminate the app.
        
        If the answer if yes than parser is destroyed and GUI terminated.

        See also
        --------
        :meth:`wiki_music.gui_lib.main_window.Checkers._exception_check`
            this method handles displaying the message to user
        :class:`wiki_music.utilities.sync.SharedVars`
            serves to pass message from parser to GUI

        Parameters
        ----------
        message: str
            message to show user when asking if app should terminate
        """
        SharedVars.ask_exit = message
        SharedVars.wait_exit = True

        while SharedVars.wait_exit:
            time.sleep(0.05)

        if SharedVars.terminate_app:
            # TODO this is not enough need to del all references to the class
            self.Preload.stop()
            del self
            sys.exit()

    # TODO if the propper page cannot be found we could show a dialog with a
    # list of possible matches for user to choose from
    def _check_band(self) -> bool:
        """Check if the artist that was input in search is the same as the one
        found on wikipedia page. If not issues warning about mismatch and asks
        user if he wants to continue.

        See also
        --------
        :meth:`terminate`
            method that takes care of ending the app execution
        """
        album_artist = self.sections["infobox"].find(
            href="/wiki/Album")  # type: ignore
        if album_artist:
            album_artist = album_artist.parent.get_text()

            if fuzz.token_set_ratio(nc(self.band), nc(album_artist)) > 90:
                return True
            else:
                b = re.sub(r"[Bb]y|[Ss]tudio album", "", album_artist).strip()
                m = (f"The Wikipedia entry for album: {self.album} belongs to "
                     f"band: {b}\nThis probably means that entry for: "
                     f"{self.album} by {self.band} does not exist.")
                self.log.exception(m)
                self.terminate(m)

                return False
        else:
            return False

    def _preload_run(self):
        """ The main method which runs in the preload thread and calls other
        methods based on input to load and parse the wikipedia page.

        See also
        --------
        :meth:`get_wiki`
            method to retrieve wikipedia page
        :meth:`cook_soup`
            method to parse the page
        """

        self.preload_running = True

        self.get_wiki(preload=True)

        if not self.error_msg:
            self.cook_soup()

            if not self.error_msg:
                self.log.info(f"Preload finished successfully, "
                                f"found: {self.url}")

        if self.error_msg:
            self.log.info(f"Preload unsucessfull: {self.error_msg}")

        self.preload_running = False

    def get_wiki(self, preload=False) -> Optional[str]:
        """ Gets wikipedia page uses offline or online version.

        Parameters
        ----------
        preload: bool
            tells the method if it is part of a running preload for GUI or not

        See also
        --------
        :attr:`wiki_music.utilities.sync.SharedVars.offline_debbug`
            defines if online version is downloaded or offline pickle version
        :meth:`_from_disk`
            fetches the offline version of page
        :meth:`_from_web`
            fetches the online version of page
        """

        # when function is called from application,
        # wait until all preloads are finished and then continue
        if not preload:
            while self.preload_running:
                time.sleep(0.05)

        try:
            downloaded = self.wiki_downloaded.get(block=False) == self.ALBUM
        except queue.Empty:
            downloaded = False
        finally:
            if downloaded:
                return self.error_msg

        self.log.debug("getting wiki")

        if SharedVars.offline_debbug:
            return self._from_disk()
        else:
            return self._from_web()

    def _from_web(self) -> Optional[str]:
        """ Guesses the right wikipedia page from innput artist and album name
        and downloads it.

        Returns
        -------
        Optional[str]
            if some error occured return string with its description
        """

        self.log.debug("from web")

        searches = [
            f"{self.album} ({self.band} album)", f"{self.album} (album)",
            self.album
        ]

        try:
            for query in searches:
                self.log.debug(f"trying query: {query}")
                self.page = wiki.page(title=query, auto_suggest=True)
                summ = nc(self.page.summary)
                if nc(self.band) in summ and nc(self.album) in summ:
                    break
            else:
                self.error_msg = "Could not get wikipedia page."

        except wiki.exceptions.DisambiguationError as e:
            print("Found entries: {}\n...".format("\n".join(e.options[:3])))
            for option in e.options:
                if self.band in option:
                    print(f"\nSelecting: {option}\n")
                    self.page = wiki.page(option)
                    break
            else:
                self.error_msg = ("Couldn't select best album entry "
                                  "from wikipedia")

        except wiki.exceptions.PageError:
            try:
                self.page = wiki.page(f"{self.album} {self.band}")
            except wiki.exceptions.PageError as e:
                self.error_msg = "Album was not found on wikipedia"

        # TODO this is dangerous, can hide other types of exceptions
        except (wiki.exceptions.HTTPTimeoutError, Exception) as e:
            self.error_msg = ("Search failed probably due to "
                              "poor internet connetion.")
        else:
            self.error_msg = None
            self.wiki_downloaded.put(self.ALBUM)

        return self.error_msg

    def _from_disk(self) -> Optional[str]:
        """Loads wikipedia page from pickle file on disk.

        Returns
        -------
        Optional[str]
            if some error occured return string with its description
        """

        # TODO pickle probably cannot handle some complex pages
        self.log.debug(f"loading pickle file {self.url}")
        if path.isfile(self.url):
            with open(self.url, 'rb') as f:
                self.log.debug("loading ...")
                self.page = pickle.load(f)
            self.log.debug("done")
            self.error_msg = None
            self.wiki_downloaded.put(self.ALBUM)
        else:
            self.error_msg = "Cannot find cached offline version of page."

        return self.error_msg

    def cook_soup(self) -> Optional[str]:
        """ Takes page downloaded from wikipedia and and parses it with use
        of bs4. Then splits the page to dictionary of sections each section
        is indexed by its name.

        Returns
        -------
        Optional[str]
            if some error occured return string with its description
        """

        try:
            cooked = self.soup_ready.get(block=False) == self.ALBUM
        except queue.Empty:
            cooked = False
        finally:
            if cooked:
                return self.error_msg

        # make BeautifulSoup black magic
        self.soup = bs4.BeautifulSoup(self.page.html(),
                                      features="lxml")  # type: ignore

        # split page to parts
        self.sections = collections.OrderedDict()

        # h2 mark the main haedings in document
        for h2 in self.soup.find_all("h2"):

            # heading should have a name marked by css class mw-headline
            # if not skip it
            name = h2.find("span", class_="mw-headline", id=True)
            if name:
                name = name["id"].lower()
            else:
                continue

            # the text belonging to each section is at the same level as
            # headings, so iterate over h2 siblings until we come across next
            # section title in document with a name
            value = []
            for s in h2.find_next_siblings():
                if s.name == "h2" and h2.find(
                        "span", class_="mw-headline", id=True):
                    break
                else:
                    value.append(s)

            self.sections[name] = value

        # add infobox to the sections
        self.sections["infobox"] = self.soup.find(
            "table", class_="infobox vevent haudio")

        # check if the album belongs to band that was requested
        if self._check_band():
            self.soup_ready.put(self.ALBUM)
            self.error_msg = None
        else:
            self.error_msg = "Album doesnt't belong to the requested band"

        return self.error_msg



""" Basic utilities used by the whole package. """

import argparse  # lazy loaded
import logging
import os
import re  # lazy loaded
import sys
import unicodedata
import webbrowser  # lazy loaded
from shutil import rmtree
from time import sleep
from typing import Any, List, Optional, Tuple, Union

import yaml  # lazy loaded

from wiki_music.constants import (API_KEY_FILE, API_KEY_MESSAGE,
                                  DONT_BOTHER_API, GOOGLE_API_URL)

from .sync import SharedVars

logging.getLogger(__name__)

__all__ = ["list_files", "to_bool", "normalize", "we_are_frozen",
           "win_naming_convetion", "flatten_set", "input_parser", "MultiLog",
           "read_google_api_key"]


class MultiLog:
    """ Passes the messages to logger instance and to SharedVars sychronization
    class where applicable as SharedVars does not implement whole Logger API

    See also
    --------
    :class:`wiki_music.utilities.sync.SharedVars`
        class passing the messages to GUI

    Parameters
    ----------
    logger: logging.Logger
        Logger instance
    """
    def __init__(self, logger):
        self.logger = logger

    def debug(self, message: Any):
        """ Issue a debug message. """
        self.logger.debug(message)

    def info(self, message: Any):
        """ Issue a info message. """
        self.logger.info(message)
        SharedVars.info(message)

    def warning(self, message: Any):
        """ Issue a warning message. """
        self.logger.warning(message)
        SharedVars.has_warning = message

    def error(self, message: Any):
        """ Issue a error message. """
        self.logger.error(message)

    def critical(self, message: Any):
        """ Issue a critical message. """
        self.logger.critical(message)

    def exception(self, message: Any):
        """ Issue a exception message. """
        self.logger.exception(message)
        SharedVars.has_exception = message


def list_files(work_dir: str, file_type: str = "music",
               recurse: bool = True) -> List[str]:
    """ List music files in directory.

    Parameters
    ----------
    work_dir: str
        directory to search
    file_type: str
        type of files to search
    recurse: bool
        whether to preform the search recursively

    Raises
    ------
    NotImplementedError
        if file_type is unsupported

    Note
    ----
    supported music formats are: .m4a, .mp3, .flac, .alac, .wav, .wma, .ogg
    supprted image formats are: .jpg, .png Qt library has trouble reading
    other formats.

    Returns
    -------
    list
        returns list of files in folder with specified file_type
    """

    found_files: List[str] = []
    allowed_types: Tuple[str, ...]

    if file_type == "music":
        # TODO list of files we ail to support:
        # (".m4a", ".mp3", ".flac", ".alac", ".wav", ".wma", ".ogg")
        allowed_types = (".m4a", ".mp3", ".flac")
    elif file_type == "image":
        allowed_types = ("jpg", "png")
    else:
        raise NotImplementedError(f"file type {file_type} is not supported")

    for root, _, files in os.walk(work_dir):
        for f in files:
            if f.endswith(allowed_types):
                found_files.append(os.path.join(root, f))
        if not recurse:
            break

    return found_files


def to_bool(string: Union[str, bool]) -> bool:
    """ Coverts string (yes, no, y, n adn capitalized versions) to bool.\n

    Parameters
    ----------
    string: str
        Input value.

    Note
    ----
    all these values and their uprercase variants are considered true:
    y, yes, t, true, <empty string>

    Returns
    -------
    bool
        truth value of given string
    """
    if isinstance(string, bool):
        return string
    else:
        return string.casefold() in ("y", "yes", "t", "true", "")


def normalize(text: str) -> str:
    """ NFKD string normalization

    Parameters
    ----------
    text: str
        text to normalize

    Returns
    -------
    str
        normalized version of text
    """
    return unicodedata.normalize("NFKD", text)


def we_are_frozen() -> bool:
    """ Checks if the running code is frozen (e.g by cx-Freeze, pyinstaller).

    Returns
    -------
    bool
        True if the code is frozen
    """

    # All of the modules are built-in to the interpreter
    return hasattr(sys, "frozen")


def read_google_api_key(GUI) -> Optional[str]:
    """ Reads google api key needed by lyricsfinder in external libraries from
    file.

    Returns
    -------
    Optional[str]
        google API key
    """

    # load google api key for lyrics search
    try:
        f = open(API_KEY_FILE, "r")
        return f.read().strip()
    except Exception:
        if not os.path.isfile(DONT_BOTHER_API):
            return _get_google_api_key(GUI)
        else:
            return None


def _get_google_api_key(GUI: bool) -> Optional[str]:
    """Prompt user to input google API key.

    Asks user through GUI or CLI if he wants to get Google API key. Three
    options are available: yes, no and don't bother me again.

    Parameters
    ----------
    GUI: bool
        if we are running in GUI or CLI mode

    Returns
    -------
    Optional[str]
        key in string format or none if key was not retrieved
    """

    # ask user if he wants to get the google API key
    if GUI:
        SharedVars.switch = "api_key"
        SharedVars.wait = True
        while SharedVars.wait:
            sleep(0.1)

        inpt = SharedVars.get_api_key
    else:
        print(API_KEY_MESSAGE)

        inpt = input("Do you want to proceed? [y(yes), n(no), "
                     "d(don't bother me again)]").lower().strip()

    if inpt == "d":
        with open(DONT_BOTHER_API, "w") as f:
            f.write("Switch file to decide if prompt for getting google API "
                    "key will be dispalyed.")
        return None
    elif to_bool(inpt):
        # open page in browser
        webbrowser.open_new_tab(GOOGLE_API_URL)

        if GUI:
            # GUI loads the key
            SharedVars.switch = "load_api_key"
            SharedVars.wait = True
            while SharedVars.wait:
                sleep(0.1)

            api_key = SharedVars.get_api_key
        else:
            # wait for key input
            api_key = str(input("Paste the key here: ")).strip()

        # write the key to file
        with open(API_KEY_FILE, "w") as f:
            f.write(api_key)

        return api_key
    else:
        return None


def win_naming_convetion(string: str, dir_name=False) -> str:
    """ Returns Windows normalized path name with removed forbiden
    characters. If platworm is not windows string is returned without changes.

    Parameters
    ----------
    string: str
        Input path to normalize
    dir_name: bool
        whether to use aditional normalization for directories

    Returns
    -------
    str
        normalized string for use in windows
    """

    if sys.platform == "nt":
        if dir_name:  # windows doesn´t like folders with dots at the end
            string = re.sub(r"\.+$", "", string)
        return re.sub(r"\<|\>|\:|\"|\/|\\|\||\?|\*", "", string)
    else:
        return string


def flatten_set(array: List[list]) -> set:
    """ Converst 2D list to 1D set.

    Parameters
    ----------
    array: list
        2D list to flatten

    Returns
    -------
    set
        1D set made of serialized lists
    """

    return set([item for sublist in array for item in sublist])


def input_parser() -> Tuple[bool, bool, bool, str, str, str, bool]:
    """Parse command line input parameters.

    Parameters
    ----------
    log: bool
        true when logger asks for argument parsing, to set the right level

    Note
    ----
    command line arguments are parsed and honoured only when running in
    CLI mode

    Returns
    -------
    bool
        yaml switch for outputing in yaml format
    bool
        offline_debug switch to turn on offline debugging
    bool
        lyrics_only search only for lyrics switch
    str
        album name
    str
        artist name
    str
        path to working directorym, default is current dir
    bool
        with_log switch on logging
    bool
        true if logging level is debug
    """

    parser = argparse.ArgumentParser(description="Description of your program")

    parser.add_argument("-y",
                        "--yaml",
                        action="store_true",
                        help="Write yaml save file?")
    parser.add_argument("-o",
                        "--offline_debbug",
                        action="store_true",
                        help="Use offline pickle debbug file instead of "
                        "web page?")
    parser.add_argument("-l",
                        "--lyrics_only",
                        action="store_true",
                        help="Download only lyrics?")
    parser.add_argument("-a",
                        "--album",
                        default=None,
                        help="Album name",
                        nargs="*")
    parser.add_argument("-b",
                        "--band",
                        default=None,
                        help="Band name",
                        nargs="*")
    parser.add_argument("-w",
                        "--work_dir",
                        default=os.getcwd(),
                        help="working directory",
                        type=str)
    parser.add_argument("-W",
                        "--With_log",
                        default=False,
                        action="store_false",
                        help="Print loggning output, "
                        "applies only to console app")
    parser.add_argument("-d",
                        "--debug",
                        default=False,
                        action="store_true",
                        help="Run in debugging mode")

    args = parser.parse_args()

    if args.album:
        args.album = " ".join(args.album)
    if args.band:
        args.band = " ".join(args.band)

    return (args.yaml, args.offline_debbug, args.lyrics_only, args.album,
            args.band, args.work_dir, args.With_log, args.debug)


def loading():
    """ CLI loading marker

    Warnings
    --------
    not implemented
    """

    # under construction
    pass



""" Module for variable synchronization between application and GUI """

import logging
from threading import Barrier, Lock
from typing import ClassVar, Union

logging.getLogger(__name__)

__all__ = ["SharedVars"]

ClsStr = ClassVar[str]
ClsBool = ClassVar[bool]


class SharedVars:
    """ Class for synchronizing info between threads SharedVars class provides
    means to synchronize some variables between gui and application,
    it serves to pass the questions asked in úarser to to PyQt GIU.
    Should not be instantiated, all methods and attributes belog to the
    class. The class implements API similar to logging.Logger with info(),
    warning() and exception() methods. These are prefered to directly
    setting the attributes value.

    Attributes
    ----------
    describe : str
        describes the current progress in parser
    ask_exit : str
        initialize dialog to exit app, works only in GUI mode
    switch : str
        decides which dialog should be displayed in GUI - lyrics, genres,
        or lyrics
    write_json : bool
        write yaml fiel contining tags for all tracks
    offline_debbug : bool
        switch to offline debugging mode. Instead of online page a local
        pickle version is used of wikipedia.Page object and cover art
        is loaded from file on local PC. Currently doesn't work for lyrics
    write_lyrics : bool
        write lyrics to tags
    assign_artists : bool
        switch telling if  we should assign artist to composers
    has_warning : str
        atttribute set by warning() method
    has_exception : str
        attribute set ba exception() method
    wait_exit : bool
        GUI wait for parser termination
    terminate_app : bool
        switch to trigger parser termination
    wait : bool
        blocks parser thread execution if true until GUI get answer from
        user
    done : bool
        anounce parser is done
    load : bool
        switch for gui to update model
    lock : Lock
        threading.Lock to sync GUI and parser thread
    barrier : Barrier
        threading.Barried  to sync GUI and parser thread
    """

    # action description
    describe: ClsStr = ""
    ask_exit: ClsStr = ""
    switch: ClsStr = ""

    # switches
    write_json: ClsBool = False
    offline_debbug: ClsBool = False
    write_lyrics: ClsBool = False
    assign_artists: ClsBool = False
    get_api_key: ClassVar[Union[str, bool]] = True

    # exceptions
    has_warning: ClsStr = ""
    has_exception: ClsStr = ""

    # control
    wait_exit: ClsBool = False
    terminate_app: ClsBool = False
    wait: ClsBool = False
    done: ClsBool = False
    load: ClsBool = False
    lock: ClassVar[Lock] = Lock()
    barrier: ClassVar[Barrier] = Barrier(2)

    @classmethod
    def re_init(cls):

        cls.write_lyrics = None
        cls.select_genre = None
        cls.assign_artists = None
        cls.write_tags = None
        cls.wait = None
        cls.switch = None
        cls.done = False
        cls.load = False

        # action description
        cls.describe = ""

        # caught exceptions
        cls.ask_exit = None

        # switches
        cls.wait_exit = False
        cls.terminate_app = False

        # threads
        cls.preload = None
        cls.parser_running = False

        # exceptions
        cls.has_warning = None
        cls.has_exception = None

    @classmethod
    def info(cls, msg: str):
        cls.lock.acquire()
        cls.describe = str(msg)
        cls.lock.release()

    @classmethod
    def warning(cls, msg: Union[Exception, str]):
        cls.has_warning = str(msg)

    @classmethod
    def exception(cls, msg: Union[Exception, str]):
        cls.has_exception = str(msg)



""" Module which holds all the important constants used throughout the package.
"""

import logging

from .colors import *
from .gui_const import *
from .parser_const import *
from .paths import *
from .tags import *

logging.getLogger(__name__)



"""Toplevel parser script that can run wikipedia search."""

import logging
from time import sleep

from wiki_music.constants.colors import CYAN, GREEN, RESET
from wiki_music.utilities import (SharedVars, exception, flatten_set, to_bool,
                                  we_are_frozen)

from .process_page import WikipediaParser

log = logging.getLogger(__name__)


class WikipediaRunner(WikipediaParser):
    """ Toplevel Wikipedia Parser class which inherits all other parser
    subclasses. This is the class that is intended for user interaction. Its
    methods know how to run the parser in order to produce meaningfull results.

    Warnings
    --------
    This is the only parser class that is ment for user interaction. Calling
    its subclasses directly might result in unexpected behaviour.

    Parameters
    ----------
    GUI: bool
        if True - assume app is running in GUI mode\n
        if False - assume app is running in CLI mode
    protected_vars: bool
        whether to initialize protected variables or not
    """
    def __init__(self, GUI: bool = True, protected_vars: bool = True) -> None:

        log.debug("init parser runner")

        super().__init__(protected_vars=protected_vars)
        self.GUI = GUI
        self.with_log = False

        log.debug("init parser runner done")

    @exception(log)
    def run_wiki(self):
        """ Runs the whole wikipedia search, together with lyrics finding. """

        if self.GUI:
            self._run_wiki_gui()
        else:
            self._run_wiki_nogui()

    def _run_wiki_gui(self):
        """ Runs wikipedia search with specifics of the GUI mode. """
        def wait_select(switch: str):
            SharedVars.switch = switch
            SharedVars.wait = True
            while SharedVars.wait:
                sleep(0.1)

        # download wikipedia page
        if not SharedVars.offline_debbug:
            self.log.info(f"Searching for: {self.album} by {self.band}")
        else:
            self.log.info("Using offline cached page insted of web page")

        error_msg = self.get_wiki()
        if error_msg:
            self.log.exception(error_msg)
            return

        else:
            self.log.info(f"Found at: {self.url}")
            self.log.info("Cooking Soup")

            error_msg = self.cook_soup()
            if error_msg:
                self.log.exception(error_msg)
                return
            else:
                self.log.info("Soup ready")

        # get page contents
        self.get_contents()

        # find release date
        self.get_release_date()
        self.log.info(f"Found release date {self.release_date}")

        # find list of genres
        self.get_genres()
        self.log.info(f"Found genre(s): {', '.join(self.genres)}")

        # download cover art from wikipedia
        self.log.info("Downloading cover art")
        self.get_cover_art()

        if not we_are_frozen():
            # basic html textout for debug
            self.basic_out()

        # print out page contents
        self.log.info(f"Found page contents: {', '.join(self.contents)}")

        # extract track list
        self.log.info("Extracting tracks")
        self._get_tracks()

        # extract personel names
        self.log.info("Extracting additional personnel")
        self.get_personnel()

        # complete artists names
        self.complete()

        # look for aditional artists in brackets behind track names and
        # complete artists names again with new info
        self.log.info("Getting tracks info")
        self.info_tracks()

        # extract writers, composers
        self.log.info("Extracting composers")
        self.get_composers()

        # complete artists names
        self.complete()

        if not we_are_frozen():
            # save to files
            self.log.info("Writing to disc")
            self.disk_write()

        # merge artists and personel which have some apperences
        self.log.info("merge artists and personnel")
        self.merge_artist_personnel()

        # select genre
        self.log.info("Select genre")
        if not self.selected_genre:
            wait_select("genres")

        # decide what to do with artists
        self.log.info("Assign artists to composers")

        # first load already known values to GUI
        # wait must be set before load,
        # otherwise if statement in gui wont be entered
        SharedVars.wait = True
        SharedVars.load = True
        while SharedVars.wait:
            sleep(0.1)

        wait_select("comp")
        if SharedVars.assign_artists:
            self.merge_artist_composers()

        # decide if you want to find lyrics
        self.log.info("Searching for Lyrics")
        wait_select("lyrics")
        self.save_lyrics()

        SharedVars.done = True
        # announce that main app thread has reached the barrier
        SharedVars.barrier.wait()

        self.log.info("Done")

    def _run_wiki_nogui(self):
        """ Runs wikipedia search with specifics of the CLI mode. """

        # download wikipedia page
        if not SharedVars.offline_debbug:

            self._log_print(msg_WHITE="Accessing Wikipedia...")

            print("Searching for: " + GREEN + self.album + RESET + " by " +
                  GREEN + self.band)

        else:
            self._log_print(msg_GREEN="Using offline cached page insted "
                            "of web page")

        error_msg = self.get_wiki()
        if error_msg:
            self._log_print(msg_GREEN=error_msg, level="WARN")
            return

        else:
            self._log_print(msg_GREEN="Found at: ", msg_WHITE=self.url)

            self._log_print(msg_WHITE="Cooking Soup")

            error_msg = self.cook_soup()
            if error_msg:
                self._log_print(msg_GREEN=error_msg, level="WARN")

                return
            else:
                self._log_print(msg_WHITE="Soup ready")

        # get page contents
        self.get_contents()

        # find release date
        self.get_release_date()
        self._log_print(msg_GREEN="Found release date",
                        msg_WHITE=self.release_date)

        # find list of genres
        self.get_genres()
        self._log_print(msg_GREEN="Found genre(s)",
                        msg_WHITE="\n".join(self.genres))

        if not we_are_frozen():
            # basic html textout for debug
            self.basic_out()

        # print out page contents
        self._log_print(msg_GREEN="Found page contents",
                        msg_WHITE="\n".join(self.contents))

        # extract track list
        self._get_tracks()

        # extract personel names
        self._log_print(msg_GREEN="Extracting additional personnel")
        self.get_personnel()

        # print out additional personnel
        self._log_print(msg_GREEN="Found aditional personel")
        if not we_are_frozen():
            print(self.personnel_2_str())

        # complete artists names
        self.complete()

        # look for aditional artists in brackets behind track names and
        # complete artists names again with new info
        self.info_tracks()

        # extract writers, composers
        self._log_print(msg_GREEN="Extracting composers")
        self.get_composers()

        # complete artists names
        self.complete()

        self._log_print(msg_GREEN="Found composers",
                        msg_WHITE="\n".join(flatten_set(self.composers)))

        if not we_are_frozen():
            # save to files
            self._log_print(msg_WHITE="Writing to disk")
            self.disk_write()

        # print out found tracklist
        self._log_print(msg_GREEN="Found Track list(s)")
        self.print_tracklist()

        # merge artists and personel which have some apperences
        self.merge_artist_personnel()

        # select genre
        if not self.selected_genre:
            if not self.genres:
                print(CYAN + "Input genre:", end="")
                self.genre = input()
            else:
                print(CYAN + "Specify which genre you want to write: [1.]")
                for i, gen in enumerate(self.genres, 1):
                    print(f"{i}. {gen}")

                print("Input number:", CYAN, end="")
                index = input()
                try:
                    index = int(index) - 1
                except ValueError:
                    index = 0

                self.selected_genre = self.genres[index]

        # decide what to do with artists
        print(CYAN + "Do you want to assign artists to composers? ([y]/n)",
              RESET,
              end=" ")
        if to_bool(input()):
            self.merge_artist_composers()

        # decide if you want to find lyrics
        print(CYAN + "\nDo you want to find and save lyrics? ([y]/n): " +
              RESET,
              end="")
        SharedVars.write_lyrics = to_bool(input())

        # download lyrics
        self.save_lyrics()

        print(CYAN + "Write data to ID3 tags? ([y]/n): " + RESET, end="")
        if to_bool(input()):
            if not self.write_tags():
                self._log_print(
                    msg_WHITE="Cannot write tags because there are no "
                    "coresponding files")
            else:
                self._log_print(msg_GREEN="Done")

    @exception(log)
    def run_lyrics(self):
        """ Runs only the lyrics search. """

        if self.GUI:
            self._run_lyrics_gui()
        else:
            self._run_lyrics_nogui()

    def _run_lyrics_gui(self):
        """ Runs only lyrics search with specifics of the GUI mode. """

        self.log.info("Searching for lyrics")

        self.save_lyrics()

        self.log.info("Done")
        SharedVars.done = True
        self.log.info("wait barrier")

        # announce that main app thread has reached the barrier
        SharedVars.barrier.wait()

    def _run_lyrics_nogui(self):
        """ Runs only lyrics search with specifics of the CLI mode. """

        self.read_files()

        # find lyrics
        self._log_print(msg_GREEN="Searching for lyrics")

        self.save_lyrics()

        if not self.write_tags():
            self._log_print(msg_WHITE="Cannot write tags because there are no "
                            "coresponding files")
        else:
            self._log_print(msg_GREEN="Done")

    def _log_print(self,
                   msg_GREEN: str = "",
                   msg_WHITE: str = "",
                   level: str = "INFO"):
        """ Redirects the input to sandard print function and to logger.

        Parameters
        ----------
        msg_GREEN: str
            message that shoul be highlighted in green in print output
        msg_WHITE: str
            message that should be left with the default font color
        level: str
            logger level for output message
        """

        if msg_GREEN != "":
            print(GREEN + "\n" + msg_GREEN)
        if msg_WHITE != "":
            print(msg_WHITE)

        if self.with_log:
            msg_GREEN = msg_GREEN + msg_WHITE

            if level == "INFO":
                log.info(msg_GREEN)
            if level == "WARN":
                log.warning(msg_GREEN)



"""Setup file for wiki_music."""

import logging
import os
import sys
from pathlib import Path
from setuptools import setup, find_packages

log = logging.getLogger(__name__)

# The directory containing this file
PKG_ROOT = os.path.dirname(__file__)

# Read package constants
with open(os.path.join(PKG_ROOT, "README.md"), "r") as f:
    README = f.read()

with open(os.path.join(PKG_ROOT, "wiki_music", "version.py"), "r") as f:
    VERSION = f.read().split(" = ")[1].replace("\"", "")

with open(os.path.join(PKG_ROOT, "requirements.txt"), "r") as f:
    REQUIREMENTS = f.read().split("\n")

# This call to setup() does all the work
setup(
    name="wiki-music",
    version=VERSION,
    description="Music tagger with information retrieval from wikipedia",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/marian-code/wikipedia-music-tags",
    author="Marián Rynik",
    author_email="marian.rynik@outlook.sk",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Environment :: X11 Applications :: Qt",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Typing :: Typed"
    ],
    packages=find_packages(exclude=("setup", "tests")),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "wiki_music_gui=wiki_music.app_gui:main",
            "wiki_music_cli=wiki_music.app_cli:main",
        ]
    },
)

# this stops the aditional data installation for readthedocs build
if os.environ.get("READTHEDOCS", False):
    sys.exit()

# optional installation of NLTK data
print("\nThe package requires downloading of NLTK data to function to its\n"
      "full potential. It will work without the data, but the extraction\n"
      "will not be as effective. Final size is ~32MB.\n")

# decide if NLTK data should be downloaded
while True:
    inpt = str(input("Do you want to proceed? [y]/n: ")).strip().casefold()

    if not inpt or inpt == "y":
        answer = True
        break
    elif inpt == "n":
        answer = False
        break
    else:
        print("You must input 'y' or 'n'")

# if yes
if answer:

    # get platform specific default path
    # get platform specific default path
    if sys.platform == "win32":
        NLTK_DEFAULT = Path("C:/nltk_data")
    elif sys.platform == "linux":
        NLTK_DEFAULT = Path("/usr/share/nltk_data")
    elif sys.platform == "darwin":
        NLTK_DEFAULT = Path("/usr/local/share/nltk_data")
    else:
        raise Warning("Usupported platform! you must specify path manually.")

    # ask user to specify their own path
    print(
        "\nPlease specify NLTK data installation path or accept the default:\n"
        " - if the path does not exist it will be created\n"
        " - if you specify other than default you wil have to modify your \n"
        "   otherwise nltk will not be able to find data folder\n")
    while True:
        inpt = str(input(f"({NLTK_DEFAULT}[ENTER]): ")).strip()

        if not inpt:
            break
        else:
            NLTK_USER = Path(inpt)

        # check for path being correct
        try:
            os.makedirs(NLTK_USER, exist_ok=True)
        except Exception:
            print("Path is not valid or cannot be created!")
        else:
            NLTK_DEFAULT = NLTK_USER
            print(f"\nAdd this entry to your system PATH variable: "
                  f"NLTK_DATA: {NLTK_DEFAULT}")
            break

    # import NLTK and download data
    import nltk

    print("\nDownloading NLTK data...\n"
          "Download size is ~90MB, Final disk size ~32MB\n")

    nltk.download("words", NLTK_DEFAULT)
    nltk.download("stopwords", NLTK_DEFAULT)
    nltk.download("maxent_ne_chunker", NLTK_DEFAULT)
    nltk.download("averaged_perceptron_tagger", NLTK_DEFAULT)
    nltk.download("punkt", NLTK_DEFAULT)

    # clean unecessary files
    print("\nDeleting unnecessary files...")
    for root, _, files in os.walk(NLTK_DEFAULT):
        for fl in files:
            if fl.endswith((".zip", ".pickle", "README")):
                # avoid deletinng some files
                if "english" not in fl and "_perceptron_tagger.p" not in fl:
                    os.remove(os.path.join(root, fl))

    # set nltk data path
    nltk.data.path.append(NLTK_DEFAULT)

    print("Done.")



"""Module which does the backgroung heavy lifting. Can obtain and parse
Wikipedia page, search and download lyrics, read and write music tags """

import logging

from .lyrics import save_lyrics
from .parser import WikipediaRunner
from .tags_io import read_tags, write_tags

__all__ = ["WikipediaRunner", "write_tags", "read_tags", "save_lyrics"]

logging.getLogger(__name__)



"""Fancy lyrics managment."""

import logging
from typing import Iterator, List

from . import extractors
from .extractor import LyricsExtractor
from .models import Lyrics, LyricsOrigin, exceptions
from .utils import UrlData, search, generate_url

log = logging.getLogger(__name__)


class LyricsManager:
    """Manage stuff."""

    extractors: List[LyricsExtractor] = []

    google_api_key = None

    @classmethod
    def setup(cls):
        log.debug("setting up")
        extractors.load_extractors()

        cls.extractors = LyricsExtractor.extractors
        log.info("loaded {} extractors".format(len(cls.extractors)))

    def extract_lyrics(self, url: str, song: str, artist: str) -> Lyrics:
        """Extract lyrics from url."""
        #log.info("extracting lyrics from url \"{}\"".format(url))
        url_data = UrlData(url)
        for ext in self.extractors:

            if not ext.can_handle(url_data):
                continue

            extractor = ext()

            log.debug("using {} for {}".format(ext, url_data))
            
            try:
                lyrics = extractor.extract_lyrics(url_data, song, artist)
            except exceptions.NoLyrics:
                log.warning(f"{extractor} didn't find any lyrics at {url}")
                continue
            except exceptions.NotAllowedError:
                log.warning(f"{extractor} couldn't access lyrics at {url}")
                continue
            except Exception:
                # switch of so it doesn´t clutter console
                # when not debugging
                log.exception(f"Something went wrong when {extractor} "
                              f"handled {url}")
                continue
            else:
                lyrics.origin = LyricsOrigin(url, extractor.name,
                                             extractor.url)
                log.debug("extracted lyrics {}".format(lyrics))
                return lyrics
        raise exceptions.NoExtractorError(url)

    def search_lyrics(self, song: str, album: str, artist: str, *,
                      google_api_key: str) -> Iterator[Lyrics]:
        """Search the net for lyrics."""
        query = artist + " " + song
        results = search(query, google_api_key)

        log.debug("got {} results fom google search".format(len(results)))

        #google search does not return any results sometimes
        if not results:
            log.warning("Google custom search returned no results! "
                        "---> Switching to url generation")
            results = generate_url(artist, album, song)

        for result in results:
            url = result["link"]

            try:
                lyrics = self.extract_lyrics(url, song, artist)
            except exceptions.NoExtractorError:
                log.warning("No extractor for url {}".format(url))
                continue
            else:
                lyrics.origin.query = query
                yield lyrics
        log.warning("No lyrics found for query \"{}\"".format(query))


LyricsManager.setup()



"""Defines exception for whole package."""

import logging

logging.getLogger(__name__)

__all__ = ["NoTracklistException", "NoReleaseDateException",
           "NoGenreException", "NoCoverArtException",
           "NoNames2ExtractException", "NoContentsException",
           "NoPersonnelException", "Mp3tagNotFoundException"]


class NoTracklistException(Exception):
    """Exception raised when no tracklist is found on wikipedia page."""

    pass


class NoReleaseDateException(Exception):
    """Exception raised when release date could not be extracted."""

    pass


class NoGenreException(Exception):
    """Exception raised when genres could not be extracted."""

    pass


class NoCoverArtException(Exception):
    """Exception raised when cover art could not be extracted."""

    pass


class NoNames2ExtractException(Exception):
    """Exception raised when cover art could not be extracted."""

    pass


class NoContentsException(Exception):
    """Exception raised when page contents could not be extracted."""

    pass


class NoPersonnelException(Exception):
    """Exception raised when page contents could not be extracted."""

    pass


class Mp3tagNotFoundException(Exception):
    """Exception raised when Mp3tag could not be run."""

    pass



"""wiki_music root file: __init__ sets which modules are lazy imported on
on package level and pulls in constants: __version__ and GUI_RUNNING."""

import logging
from sys import argv

from lazy_import import lazy_module

logging.getLogger(__name__)

try:
    from version import __version__
except ImportError:
    __version__ = "unknown"  #: specifies version of the package

if "gui" in argv[0].lower():
    GUI_RUNNING = True  #: tells if GUI or CLI frontend is running
else:
    GUI_RUNNING = False

# mark modules as lazy
# if we are subsequently importing submodule, it must be done as:
# import module.sub_module as sub_module
lazy_module("re")
lazy_module("io")
lazy_module("bs4")
lazy_module("yaml")
lazy_module("queue")
lazy_module("pickle")
lazy_module("urllib")
lazy_module("winreg")
lazy_module("mutagen")
lazy_module("requests")
lazy_module("argparse")
lazy_module("wikipedia")
lazy_module("webbrowser")
lazy_module("subprocess")
lazy_module("datefinder")
lazy_module("collections")
lazy_module("PIL.Image")
lazy_module("PIL.ImageFile")
lazy_module("fuzzywuzzy.fuzz")
lazy_module("fuzzywuzzy.process")
lazy_module("wiki_music.external_libraries.lyricsfinder")
lazy_module("wiki_music.external_libraries.google_images_download"
            ".google_images_download")
lazy_module("wiki_music.external_libraries.google_images_download"
            ".google_images_download_offline")



"""Extractor for lyricsmode.com."""

import logging

from ..extractor import LyricsExtractor
from ..models.lyrics import Lyrics
from ..models import exceptions

log = logging.getLogger(__name__)


class Lyricsmode(LyricsExtractor):
    """Class for extracting lyrics."""

    name = "Lyricsmode"
    url = "http://www.lyricsmode.com/"
    display_url = "lyricsmode.com"

    @staticmethod
    def extract_lyrics(url_data, song, artist):
        """Extract lyrics."""
        bs = url_data.bs

        # 1st possible lyrics location
        lyrics_window = bs.find("p", {"id": "lyrics_text",
                                      "class": "ui-annotatable"})

        # 2nd possible lyrics location
        if not lyrics_window:
            lyrics_window = bs.find("div", {"id": "lyrics_text"})

        # check for 404 ERROR page not found
        if not lyrics_window:
            if "oops" in bs.find("div", {"class":
                                         "song_name fs32"}).text.lower():
                raise exceptions.NoLyrics

        lyrics = lyrics_window.text

        title = bs.find("h1", attrs={"class": "song_name fs32"}).text[:-7]

        return Lyrics(title, lyrics)



"""Extractor for lyrical-nonsense.com."""

import logging

from .. import utils
from ..extractor import LyricsExtractor
from ..models.lyrics import Lyrics
from ..utils import UrlData

log = logging.getLogger(__name__)


class LyricalNonsense(LyricsExtractor):
    """Class for extracting lyrics."""

    name = "Lyrical Nonsense"
    url = "http://www.lyrical-nonsense.com/"
    display_url = "lyrical-nonsense.com"

    @staticmethod
    def extract_lyrics(url_data: UrlData, song, artist) -> Lyrics:
        """Extract lyrics."""
        bs = url_data.bs
        title = bs.select_one("div.titletext2new h3").text
        artist = bs.select_one("div.artisttext1new h2 a").text

        lyrics_window = bs.select_one("div#Romaji div.olyrictext") or bs.select_one("div#Lyrics div.olyrictext")
        lyrics = utils.clean_lyrics(lyrics_window.text)

        return Lyrics(title, lyrics, artist=artist)



"""Utility functions used by GUI classes."""

import io  # lazy loaded
import logging
import os
import sys
import urllib  # lazy loaded
import winreg  # lazy loaded
from typing import Optional, Tuple

import requests  # lazy loaded
from PIL import Image, ImageFile  # lazy loaded

from wiki_music.constants.paths import FILES_DIR

logging.getLogger(__name__)

# TODO meditate on including clipboard interaction
# import win32clipboard  # lazy loaded


__all__ = ["get_music_path", "abstract_warning", "get_image", "get_sizes",
           "comp_res", "get_image_size", "get_icon"]  # , "send_to_clipboard"]


def abstract_warning():
    """ Raises error when abstract method is called directly.
    Raises
    ------
    NotImplementedError
        this is the sole purpose of this function
    """
    raise NotImplementedError("This method is abstaract and should be "
                              "reimplemented by inheriting class")


# TODO not used for now, need to rethink, maybe it is useless when we have
# writing to tags, it has a complex dependency which does not play nice with
# other platform except windows: pywin32>=224; platform_system == "Windows"
# maybe this function sholud be removed.
# ! in the future it might be replaced by PyQt QClipboard
# ! https://www.tutorialspoint.com/pyqt/pyqt_qclipboard.htm
"""
def send_to_clipboard(data: bytearray, clip_type: Optional[int] = None):
     Pastes data to clipboard.

    Parameters
    ----------
    clip_type: int
        type of the clipboard
    data: bytearray
        data to paste to clipboard

    Raises
    ------
    NotImplementedError
        when unsupported platform is detected, and we don't know how to
        interact with cilpboard

    clip_data: bytes

    file_stream = io.BytesIO(data)
    image = Image.open(file_stream)

    clip_stream = io.BytesIO()
    image.convert("RGB").save(clip_stream, "BMP")
    clip_data = clip_stream.getvalue()[14:]
    clip_stream.close()

    if os.name == "nt":
        if not clip_type:
            clip_type = win32clipboard.CF_DIB

        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(clip_type, clip_data)
        win32clipboard.CloseClipboard()
    else:
        raise NotImplementedError("Only Windows clipboard is supported "
                                  "at the moment")
"""


def get_music_path() -> str:
    """Returns the default music path for linux or windows

    Returns
    -------
    str
        string path pointing to music library location
    """

    if os.name == "nt":
        sub_key = (r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer"
                   r"\Shell Folders")
        music_guid = "{4BD8D571-6D19-48D3-BE97-422220080E43}"
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
                location = winreg.QueryValueEx(key, music_guid)[0]
            return location
        except Exception:
            return "~"

    else:
        return os.path.join(os.path.expanduser("~"), "music")


def get_image(address: str) -> Optional[bytes]:
    """ Based on addres decides if the image is online or local. If address
    has http prefix, image is downloaded from internet. If not then it is read
    from disk.

    Parameters
    ----------
    address:
        string with path to picture on local PC or with http address

    Returns
    -------
    bytes
        bytes repesentation of image loaded to memory or None if address is
        not valid
    """

    if address.startswith("http"):
        return requests.get(address).content
    else:
        # this is for offline debug
        try:
            with open(address, "rb") as f:
                return bytearray(f.read())
        except Exception as e:
            print(f"Address: {address} could not be opened in "
                  f"online or offline mode. {e}")
            return None


def comp_res(image: bytearray, quality: int, x: int = 0, y: int = 0) -> bytes:
    """ Compress and/or change image resolution. If x and y dimension are
    both specified than image is resized to these dimension otherwise it is
    only compressed

    Parameters
    ----------
    image: bytearray
        bytes representation of image loaded to memory
    quality: int
        target quality to which we wanto compress the limits are (1, 99)
    x: int
        horizontal image dimension
    y: int
        vertical iage dimension

    Returns
    -------
    bytes
        bytes image compressed and resized than reloaded to memory
    """

    FORMAT = Image.registered_extensions()[".jpg"]

    file_stream = io.BytesIO(image)
    img = Image.open(file_stream)

    if x and y:
        img = img.resize((x, y), Image.LANCZOS)

    file_stream = io.BytesIO()
    img.save(file_stream,
             FORMAT,
             optimize=True,
             quality=quality,
             progressive=True)

    return file_stream.getvalue()


def get_image_size(image: bytearray) -> str:
    """ get size of image in memory

    Parameters
    ----------
    image: bytearray
        bytes image in loaded in memory

    Returns
    -------
    str
        string with image size in Kb rounded to 2 decimal places
    """

    return f"{sys.getsizeof(image) / 1024:.2f}"


def get_icon() -> str:
    """ returns application icon path

    Raises
    ------
    FileNotFoundError
        if the path does not exist

    Returns
    -------
    str
        string with icon path
    """

    icon = os.path.join(FILES_DIR, "icon.ico")
    if not os.path.isfile(icon):
        raise FileNotFoundError(f"There is no icon file in: {icon}")
    else:
        return icon


def get_sizes(uri: str) -> Tuple[Optional[int], Optional[Tuple[int, int]]]:
    """ Get file size and image size (None if not known) of picture on the
    internet without downloading it.

    Parameters
    ----------
    uri: str
        picture url addres

    References
    ----------
    https://stackoverflow.com/questions/15800704/get-image-size-without-loading-image-into-memory

    Returns
    -------
    tuple
        if the size can be obtained result is a tuple with picture size in Kb
        and dimensions tuple as a second element
    """

    try:
        fl = urllib.request.urlopen(uri)
        size = fl.headers.get("content-length")
        if size:
            size = int(size)

        p = ImageFile.Parser()
        while True:
            data = fl.read(1024)
            if not data:
                break
            p.feed(data)
            if p.image:
                return size, p.image.size

        fl.close()

        return size, None
    except Exception:
        return None, None
